@@ -20,7 +20,7 @@
     '''
 import platform
 
-__version__ = '21.12.0'
+__version__ = '21.17.0'
 
 import tkinter
 import tkinter.ttk
@@ -188,4 +188,4 @@ class ControlPanelButton(tkinter.Button):
 
     def takeSnapshot(self):
         # No need to retake snapshot as it is already shown
-        self.culebron.saveSnapshot()
+        self.culebron.saveSnapshot(showDialog=True)
