@@ -20,7 +20,7 @@ limitations under the License.
 
 from __future__ import print_function
 
-__version__ = '21.12.0'
+__version__ = '21.17.0'
 
 import json
 import os
@@ -32,10 +32,11 @@ import threading
 import time
 import warnings
 from abc import ABC
-from typing import Optional
+from datetime import datetime
+from typing import Optional, List, Tuple
 
 import culebratester_client
-from culebratester_client import Text, ObjectRef, StatusResponse
+from culebratester_client import Text, ObjectRef, StatusResponse, DefaultApi, Point, PerformTwoPointerGestureBody
 
 from com.dtmilano.android.adb.adbclient import AdbClient
 from com.dtmilano.android.common import obtainAdbPath
@@ -120,7 +121,7 @@ class UiAutomatorHelper:
         self.osName = platform.system()
         ''' The OS name. We sometimes need specific behavior. '''
         self.isDarwin = (self.osName == 'Darwin')
-        ''' Is it Mac OSX? '''
+        ''' Is it macOS? '''
         self.hostname = hostname
         ''' The hostname we are connecting to. '''
 
@@ -129,11 +130,12 @@ class UiAutomatorHelper:
             file=sys.stderr)
         configuration = culebratester_client.Configuration()
         configuration.host = f'http://{hostname}:{localport}/{api_version}'
-        self.api_instance = culebratester_client.DefaultApi(culebratester_client.ApiClient(configuration))
+        self.api_instance: DefaultApi = culebratester_client.DefaultApi(culebratester_client.ApiClient(configuration))
         self.device: UiAutomatorHelper.Device = UiAutomatorHelper.Device(self)
         self.target_context: UiAutomatorHelper.TargetContext = UiAutomatorHelper.TargetContext(self)
         self.object_store: UiAutomatorHelper.ObjectStore = UiAutomatorHelper.ObjectStore(self)
         self.ui_device: UiAutomatorHelper.UiDevice = UiAutomatorHelper.UiDevice(self)
+        self.ui_object: UiAutomatorHelper.UiObject = UiAutomatorHelper.UiObject(self)
         self.ui_object2: UiAutomatorHelper.UiObject2 = UiAutomatorHelper.UiObject2(self)
         self.until: UiAutomatorHelper.Until = UiAutomatorHelper.Until(self)
 
@@ -201,6 +203,18 @@ class UiAutomatorHelper:
     def __httpCommand(self, url, params=None, method='GET'):
         raise RuntimeError(f"this method should not be used: url={url} params={params} method={method}")
 
+    @staticmethod
+    def timestamp() -> str:
+        """
+        Timestamp in ISO format.
+
+        WARNING: may not be suitable to include in filenames on some platforms as it may include invalid path chars
+        (i.e. `:`).
+
+        :return: the timestamps
+        """
+        return datetime.now().isoformat()
+
     #
     # UiAutomatorHelper internal commands
     #
@@ -302,6 +316,7 @@ class UiAutomatorHelper:
         :deprecated: use uiAutomatorHelper.device.display_real_size()
         :return: the display real size
         """
+        warnings.warn("Deprecated: will be removed in future versions")
         return self.api_instance.device_display_real_size_get()
 
     #
@@ -373,20 +388,14 @@ class UiAutomatorHelper:
         def __init__(self, uiAutomatorHelper) -> None:
             super().__init__(uiAutomatorHelper)
 
-        def click(self, **kwargs):
+        def click(self, x: int, y: int):
             """
-            Clicks.
+            Clicks on the specified coordinates.
 
             :see https://github.com/dtmilano/CulebraTester2-public/blob/master/openapi.yaml
-            :param kwargs:
             :return:
             """
-            if self.all(['x', 'y'], kwargs):
-                x = int(kwargs['x'])
-                y = int(kwargs['y'])
-                return self.uiAutomatorHelper.api_instance.ui_device_click_get(x=x, y=y)
-            else:
-                raise ValueError('click: (x, y) must have a value')
+            return self.uiAutomatorHelper.api_instance.ui_device_click_get(x=x, y=y)
 
         def dump_window_hierarchy(self, _format='JSON'):
             """
@@ -430,6 +439,22 @@ class UiAutomatorHelper:
             body = culebratester_client.Selector(**kwargs)
             return self.uiAutomatorHelper.api_instance.ui_device_find_objects_post(body=body)
 
+        def has_object(self, **kwargs) -> bool:
+            """
+            Has an object.
+            Invokes GET or POST method depending on the arguments passed.
+
+            :see https://github.com/dtmilano/CulebraTester2-public/blob/master/openapi.yaml
+            :param kwargs:
+            :return: True or False
+            """
+            if 'body' in kwargs:
+                return self.uiAutomatorHelper.api_instance.ui_device_has_object_post(**kwargs).value
+            if self.some(['resource_id', 'ui_selector', 'by_selector'], kwargs):
+                return self.uiAutomatorHelper.api_instance.ui_device_has_object_get(**kwargs).value
+            body = culebratester_client.Selector(**kwargs)
+            return self.uiAutomatorHelper.api_instance.ui_device_has_object_post(body=body).value
+
         def press_back(self):
             """
             Presses BACK.
@@ -457,6 +482,15 @@ class UiAutomatorHelper:
             """
             return self.uiAutomatorHelper.api_instance.ui_device_press_home_get()
 
+        def press_recent_apps(self):
+            """
+            Press recent apps.
+
+            :see https://github.com/dtmilano/CulebraTester2-public/blob/master/openapi.yaml
+            :return:
+            """
+            return self.uiAutomatorHelper.api_instance.ui_device_press_recent_apps_get()
+
         def press_key_code(self, key_code, meta_state=0):
             """
             Presses a key code.
@@ -538,6 +572,68 @@ class UiAutomatorHelper:
             """
             return self.uiAutomatorHelper.api_instance.ui_device_wait_for_window_update_get(timeout, **kwargs)
 
+    #
+    # UiObject
+    #
+    class UiObject(ApiBase):
+        """
+        Inner UiObject class.
+
+        Notice that this class does not require an OID in its constructor.
+        """
+
+        def __init__(self, uiAutomatorHelper) -> None:
+            super().__init__(uiAutomatorHelper)
+
+        def performTwoPointerGesture(self, oid: int, startPoint1: Tuple[int, int], startPoint2: Tuple[int, int],
+                                     endPoint1: Tuple[int, int], endPoint2: Tuple[int, int], steps: int):
+            """
+            Generates a two-pointer gesture with arbitrary starting and ending points.
+            :see https://github.com/dtmilano/CulebraTester2-public/blob/master/openapi.yaml
+            :param oid: the oid
+            :param startPoint1:
+            :param startPoint2:
+            :param endPoint1:
+            :param endPoint2:
+            :param steps:
+            :return: the result of the operation
+            """
+            body = PerformTwoPointerGestureBody(Point(startPoint1[0], startPoint1[1]),
+                                                Point(startPoint2[0], startPoint2[1]),
+                                                Point(endPoint1[0], endPoint1[1]),
+                                                Point(endPoint2[0], endPoint2[1]),
+                                                steps)
+            return self.uiAutomatorHelper.api_instance.ui_object_oid_perform_two_pointer_gesture_post(oid=oid,
+                                                                                                      body=body)
+
+        def pinch_in(self, oid: int, percentage: int, steps: int = 50):
+            """
+            Performs a two-pointer gesture, where each pointer moves diagonally toward the other, from the edges to the
+            center of this UiObject.
+            :see https://github.com/dtmilano/CulebraTester2-public/blob/master/openapi.yaml
+            :param oid: the oid
+            :param percentage: percentage of the object's diagonal length for the pinch gesture
+            :param steps: the number of steps for the gesture. Steps are injected about 5 milliseconds apart, so 100
+            steps may take around 0.5 seconds to complete.
+            :return: the result of the operation
+            """
+            return self.uiAutomatorHelper.api_instance.ui_object_oid_pinch_in_get(oid=oid, percentage=percentage,
+                                                                                  steps=steps)
+
+        def pinch_out(self, oid: int, percentage: int, steps: int = 50):
+            """
+            Performs a two-pointer gesture, where each pointer moves diagonally opposite across the other, from the
+            center out towards the edges of the this UiObject.
+            :see https://github.com/dtmilano/CulebraTester2-public/blob/master/openapi.yaml
+            :param oid: the oid
+            :param percentage: percentage of the object's diagonal length for the pinch gesture
+            :param steps: the number of steps for the gesture. Steps are injected about 5 milliseconds apart, so 100
+            steps may take around 0.5 seconds to complete.
+            :return: the result of the operation
+            """
+            return self.uiAutomatorHelper.api_instance.ui_object_oid_pinch_out_get(oid=oid, percentage=percentage,
+                                                                                   steps=steps)
+
     #
     # UiObject2
     #
@@ -551,7 +647,7 @@ class UiAutomatorHelper:
         def __init__(self, uiAutomatorHelper) -> None:
             super().__init__(uiAutomatorHelper)
 
-        def clear(self, oid):
+        def clear(self, oid: int):
             """
             Clears the text content if this object is an editable field.
             :see https://github.com/dtmilano/CulebraTester2-public/blob/master/openapi.yaml
@@ -560,7 +656,7 @@ class UiAutomatorHelper:
             """
             return self.uiAutomatorHelper.api_instance.ui_object2_clear_get(oid=oid)
 
-        def click(self, oid):
+        def click(self, oid: int):
             """
             Clicks on this object.
             :see https://github.com/dtmilano/CulebraTester2-public/blob/master/openapi.yaml
@@ -649,7 +745,7 @@ class UiAutomatorHelper:
             body = culebratester_client.Selector(**kwargs)
             return self.uiAutomatorHelper.api_instance.until_find_object_post(body=body)
 
-        def find_objects(self, **kwargs) -> ObjectRef:
+        def find_objects(self, **kwargs) -> List[ObjectRef]:
             """
             Returns a SearchCondition that is satisfied when at least one element matching the selector can be found.
             The condition will return the first matching element.
@@ -678,6 +774,7 @@ class UiAutomatorHelper:
         :param kwargs:
         :return:
         """
+        warnings.warn("Deprecated: will be removed in future versions")
         params = kwargs
         if not (('x' in params and 'y' in params) or 'oid' in params):
             raise RuntimeError('click: (x, y) or oid must have a value')
@@ -693,24 +790,28 @@ class UiAutomatorHelper:
         """
         :deprecated:
         """
+        warnings.warn("Deprecated: will be removed in future versions")
         return self.api_instance.ui_device_dump_window_hierarchy_get(format='JSON')
 
     def findObject(self, **kwargs):
         """
         :deprecated:
         """
+        warnings.warn("Deprecated: will be removed in future versions")
         return self.api_instance.ui_device_find_object_get(**kwargs)
 
     def findObjects(self, **kwargs):
         """
         :deprecated:
         """
+        warnings.warn("Deprecated: will be removed in future versions")
         return self.api_instance.ui_device_find_objects_get(**kwargs)
 
     def longClick(self, **kwargs):
         """
         :deprecated:
         """
+        warnings.warn("Deprecated: will be removed in future versions")
         params = kwargs
         if not (('x' in params and 'y' in params) or 'oid' in params):
             raise RuntimeError('longClick: (x, y) or oid must have a value')
@@ -724,42 +825,49 @@ class UiAutomatorHelper:
         """
         :deprecated:
         """
+        warnings.warn("Deprecated: will be removed in future versions")
         return self.__httpCommand('/UiDevice/openNotification')
 
     def openQuickSettings(self):
         """
         :deprecated:
         """
+        warnings.warn("Deprecated: will be removed in future versions")
         return self.__httpCommand('/UiDevice/openQuickSettings')
 
     def pressBack(self):
         """
         :deprecated:
         """
+        warnings.warn("Deprecated: will be removed in future versions")
         return self.api_instance.ui_device_press_back_get()
 
     def pressHome(self):
         """
         :deprecated:
         """
+        warnings.warn("Deprecated: will be removed in future versions")
         return self.api_instance.ui_device_press_home_get()
 
     def pressKeyCode(self, keyCode, metaState=0):
         """
         :deprecated:
         """
+        warnings.warn("Deprecated: will be removed in future versions")
         return self.api_instance.ui_device_press_key_code_get(key_code=keyCode, meta_state=metaState)
 
     def pressRecentApps(self):
         """
         :deprecated:
         """
+        warnings.warn("Deprecated: will be removed in future versions")
         return self.__httpCommand('/UiDevice/pressRecentApps')
 
     def swipe(self, startX=-1, startY=-1, endX=-1, endY=-1, steps=10, segments=None, segmentSteps=5):
         """
         :deprecated:
         """
+        warnings.warn("Deprecated: will be removed in future versions")
         if segments is None:
             segments = []
         if startX != -1 and startY != -1:
@@ -776,12 +884,14 @@ class UiAutomatorHelper:
         """
         :deprecated:
         """
+        warnings.warn("Deprecated: will be removed in future versions")
         return self.api_instance.ui_device_screenshot_get(scale=scale, quality=quality, **kwargs)
 
     def waitForIdle(self, timeout):
         """
         :deprecated:
         """
+        warnings.warn("Deprecated: will be removed in future versions")
         params = {'timeout': timeout}
         return self.api_instance.ui_device_wait_for_idle_get(**params)
 
@@ -792,6 +902,7 @@ class UiAutomatorHelper:
         """
         :deprecated:
         """
+        warnings.warn("Deprecated: will be removed in future versions")
         body = {'text': text}
         return self.api_instance.ui_object2_oid_set_text_post(body, oid)
 
@@ -804,10 +915,12 @@ class UiAutomatorHelper:
         :param timeout:
         :return:
         """
+        warnings.warn("Deprecated: will be removed in future versions")
         params = {'eventCondition': eventCondition, 'timeout': timeout}
         return self.__httpCommand('/UiObject2/%d/clickAndWait' % uiObject2.oid, params)
 
     def getText(self, oid):
+        warnings.warn("Deprecated: will be removed in future versions")
         return self.api_instance.ui_object2_oid_get_text_get(oid)
 
     def isChecked(self, uiObject=None):
@@ -817,6 +930,7 @@ class UiAutomatorHelper:
         :param uiObject:
         :return:
         """
+        warnings.warn("Deprecated: will be removed in future versions")
         # This path works for UiObject and UiObject2, so there's no need to handle both cases differently
         path = '/UiObject/%d/isChecked' % uiObject.oid
         response = self.__httpCommand(path, None)
@@ -829,6 +943,7 @@ class UiAutomatorHelper:
     # UiScrollable
     #
     def uiScrollable(self, path, params=None):
+        warnings.warn("Deprecated: will be removed in future versions")
         response = self.__httpCommand('/UiScrollable/' + path, params)
         if DEBUG:
             print("UiAutomatorHelper: uiScrollable: response=", response, file=sys.stderr)
@@ -857,26 +972,33 @@ class UiObject:
     """
 
     def __init__(self, uiAutomatorHelper, oid, response):
+        warnings.warn("Deprecated: will be removed in future versions")
         self.uiAutomatorHelper = uiAutomatorHelper
         self.oid = oid
         self.className = response['className']
 
     def getOid(self):
+        warnings.warn("Deprecated: will be removed in future versions")
         return self.oid
 
     def getClassName(self):
+        warnings.warn("Deprecated: will be removed in future versions")
         return self.className
 
     def click(self):
+        warnings.warn("Deprecated: will be removed in future versions")
         self.uiAutomatorHelper.click(oid=self.oid)
 
     def longClick(self):
+        warnings.warn("Deprecated: will be removed in future versions")
         self.uiAutomatorHelper.longClick(oid=self.oid)
 
     def getText(self):
+        warnings.warn("Deprecated: will be removed in future versions")
         return self.uiAutomatorHelper.getText(uiObject=self)
 
     def setText(self, text):
+        warnings.warn("Deprecated: will be removed in future versions")
         self.uiAutomatorHelper.setText(uiObject=self, text=text)
 
 
@@ -896,6 +1018,7 @@ class UiObject2:
         :param class_name: the class name of the UI object
         :param oid: the object id
         """
+        warnings.warn("Deprecated: will be removed in future versions")
         self.uiAutomatorHelper = uiAutomatorHelper
         self.class_name = class_name
         self.oid = oid
@@ -906,6 +1029,7 @@ class UiObject2:
         :deprecated:
         :return: the result of the operation
         """
+        warnings.warn("Deprecated: will be removed in future versions")
         return self.uiAutomatorHelper.ui_object2.clear(oid=self.oid)
 
     def click(self):
@@ -914,6 +1038,7 @@ class UiObject2:
         :deprecated:
         :return: the result of the operation
         """
+        warnings.warn("Deprecated: will be removed in future versions")
         return self.uiAutomatorHelper.ui_object2.click(oid=self.oid)
 
     def dump(self):
@@ -922,6 +1047,7 @@ class UiObject2:
         :deprecated:
         :return: the content of the object
         """
+        warnings.warn("Deprecated: will be removed in future versions")
         return self.uiAutomatorHelper.ui_object2.dump(oid=self.oid)
 
     def get_text(self):
@@ -930,6 +1056,7 @@ class UiObject2:
         Returns the text value for this object.
         :return: the text
         """
+        warnings.warn("Deprecated: will be removed in future versions")
         return self.uiAutomatorHelper.ui_object2.get_text(oid=self.oid)
 
     def long_click(self):
@@ -938,6 +1065,7 @@ class UiObject2:
         :deprecated:
         :return: the result of the operation
         """
+        warnings.warn("Deprecated: will be removed in future versions")
         return self.uiAutomatorHelper.ui_object2.long_click(oid=self.oid)
 
     def set_text(self, text):
@@ -947,6 +1075,7 @@ class UiObject2:
         :param text: the text
         :return: the result of the operation
         """
+        warnings.warn("Deprecated: will be removed in future versions")
         return self.uiAutomatorHelper.ui_object2.set_text(oid=self.oid, text=text)
 
     def clickAndWait(self, eventCondition, timeout):
@@ -956,6 +1085,7 @@ class UiObject2:
         :param timeout:
         :return:
         """
+        warnings.warn("Deprecated: will be removed in future versions")
         self.uiAutomatorHelper.clickAndWait(uiObject2=self, eventCondition=eventCondition, timeout=timeout)
 
     def isChecked(self):
@@ -964,6 +1094,7 @@ class UiObject2:
         :deprecated:
         :rtype: bool
         """
+        warnings.warn("Deprecated: will be removed in future versions")
         return self.uiAutomatorHelper.isChecked(uiObject=self)
 
     def longClick(self):
@@ -971,6 +1102,7 @@ class UiObject2:
 
         :deprecated:
         """
+        warnings.warn("Deprecated: will be removed in future versions")
         self.uiAutomatorHelper.longClick(oid=self.oid)
 
     def getText(self):
@@ -979,6 +1111,7 @@ class UiObject2:
         :deprecated:
         :return:
         """
+        warnings.warn("Deprecated: will be removed in future versions")
         # NOTICE: even if this is an uiObject2 we are invoking the only "getText" method in UiAutomatorHelper
         # passing the uiObject2 as uiObject
         return self.uiAutomatorHelper.getText(uiObject=self)
@@ -989,6 +1122,7 @@ class UiObject2:
         :deprecated:
         :param text:
         """
+        warnings.warn("Deprecated: will be removed in future versions")
         # NOTICE: even if this is an uiObject2 we are invoking the only "setText" method in UiAutomatorHelper
         # passing the uiObject2 as uiObject
         self.uiAutomatorHelper.setText(uiObject=self, text=text)
@@ -996,26 +1130,33 @@ class UiObject2:
 
 class UiScrollable:
     def __init__(self, uiAutomatorHelper, uiSelector):
+        warnings.warn("Deprecated: will be removed in future versions")
         self.uiAutomatorHelper = uiAutomatorHelper
         self.uiSelector = uiSelector
         self.oid, self.response = self.__createUiScrollable()
 
     def __createUiScrollable(self):
+        warnings.warn("Deprecated: will be removed in future versions")
         return self.uiAutomatorHelper.uiScrollable('new', {'uiSelector': self.uiSelector})
 
     def flingBackward(self):
+        warnings.warn("Deprecated: will be removed in future versions")
         return self.uiAutomatorHelper.uiScrollable(str(self.oid) + '/flingBackward')
 
     def flingForward(self):
+        warnings.warn("Deprecated: will be removed in future versions")
         return self.uiAutomatorHelper.uiScrollable(str(self.oid) + '/flingForward')
 
     def flingToBeginning(self, maxSwipes=20):
+        warnings.warn("Deprecated: will be removed in future versions")
         return self.uiAutomatorHelper.uiScrollable(str(self.oid) + '/flingToBeginning', {'maxSwipes': maxSwipes})
 
     def flingToEnd(self, maxSwipes=20):
+        warnings.warn("Deprecated: will be removed in future versions")
         return self.uiAutomatorHelper.uiScrollable(str(self.oid) + '/flingToEnd', {'maxSwipes': maxSwipes})
 
     def getChildByDescription(self, uiSelector, description, allowScrollSearch):
+        warnings.warn("Deprecated: will be removed in future versions")
         oid, response = self.uiAutomatorHelper.uiScrollable(str(self.oid) + '/getChildByDescription',
                                                             {'uiSelector': uiSelector,
                                                              'contentDescription': description,
@@ -1023,15 +1164,18 @@ class UiScrollable:
         return UiObject(self.uiAutomatorHelper, oid, response)
 
     def getChildByText(self, uiSelector, text, allowScrollSearch):
+        warnings.warn("Deprecated: will be removed in future versions")
         oid, response = self.uiAutomatorHelper.uiScrollable(str(self.oid) + '/getChildByText',
                                                             {'uiSelector': uiSelector, 'text': text,
                                                              'allowScrollSearch': allowScrollSearch})
         return UiObject(self.uiAutomatorHelper, oid, response)
 
     def setAsHorizontalList(self):
+        warnings.warn("Deprecated: will be removed in future versions")
         self.uiAutomatorHelper.uiScrollable(str(self.oid) + '/setAsHorizontalList')
         return self
 
     def setAsVerticalList(self):
+        warnings.warn("Deprecated: will be removed in future versions")
         self.uiAutomatorHelper.uiScrollable(str(self.oid) + '/setAsVerticalList')
         return self
